#!/usr/bin/python
# Sampling pointcloud and running inference of dgcnn

from __future__ import print_function
import os
import argparse
import torch
import torch.nn as nn
import matplotlib.pyplot as plt
import numpy as np
from config.config import *
from lib.util import IOStream
from lib.model import DGCNN
# load args
args = parse_args()

device = torch.device("cuda" if args.cuda else "cpu")


def test(args, data, visualize=True):
		#Try to load models
		model = DGCNN(args).to(device)
		model = nn.DataParallel(model)
		model.load_state_dict(torch.load(args.model_path))
		model = model.eval()
		####################
		# Testing
		####################
		model.eval()

		data1 = data.permute(0, 2, 1)
		preds = model(data1)
		preds = preds.reshape((preds.shape[0], args.num_kpts, 3)) #reshape to (8,3) keypoints
		## back to numpy from
		preds = preds.detach().cpu().numpy()
		data = data.detach().cpu().numpy()
		if visualize:
				plt.figure(figsize=(20,10))
				ax = plt.axes(projection='3d')
				ax.set_xlim([-3, 3])
				ax.set_ylim([-3, 3])
				ax.set_zlim([0, 3])
				ax.set_axis_off()
				## show mesh
				ax.scatter(data[0,:,0], data[0,:,1], data[0,:,2], s=1, color='gray')
				ax.scatter(preds[0,:,0], preds[0,:,1], preds[0,:,2], s=3, color='red')
				# np.save("/home/srikanth/pose3d/geometric/test2_car_downsampled.npy", data[0,:,:])
				# np.save("/home/srikanth/pose3d/geometric/test2_car_kpts.npy", preds[0,:,:])
				plt.show()

if __name__ == "__main__":

		args.cuda = args.cuda and torch.cuda.is_available()
		torch.manual_seed(args.seed)
		if args.cuda:
				print('Using GPU : ' + str(torch.cuda.current_device()) + ' from ' + str(torch.cuda.device_count()) + ' devices')
				torch.cuda.manual_seed(args.seed)
		else:
				print('Using CPU')

		# manual loading of a pointcloud to test
		carPC = np.load("/home/srikanth/pose3d/geometric/test2_car.npy")

		## uniformly sample 1024 points from the pointcloud
		carPC = carPC[np.argsort(carPC[:, 2])]
		## random selection of points
		random_indices = np.random.choice(carPC.shape[0], size=args.num_points, replace=False)
		carPC = carPC[random_indices, :]
		## fixed selection
		carPC = carPC[0:1024,:]
		carPC = np.expand_dims(carPC, axis=0)
		carPC = torch.Tensor(carPC).to(device)
		test(args, carPC)
