#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import sys
import glob
import h5py
import numpy as np
from torch.utils.data import Dataset
from transforms3d.euler import euler2mat
from numpy import linalg as LA
import open3d as o3d
import torch
import scipy.spatial as sp
import matplotlib as mpl
import matplotlib.cm as cm
import copy
from lib.util import dirac_sampling

def download(DATA_DIR):
		BASE_DIR = os.path.dirname(os.path.abspath(__file__))
		# DATA_DIR = os.path.join(BASE_DIR, '../data')
		DOWNLOAD_SCRIPT = os.path.join(BASE_DIR, '../download_data.sh')
		os.system('bash %s'%(DOWNLOAD_SCRIPT))
		# if not os.path.exists(DATA_DIR):
		#     os.mkdir(DATA_DIR)
		# if not os.path.exists(os.path.join(DATA_DIR, 'modelnet40_ply_hdf5_2048')):
		#     www = 'https://shapenet.cs.stanford.edu/media/modelnet40_ply_hdf5_2048.zip'
		#     zipfile = os.path.basename(www)
		#     os.system('wget %s; unzip %s' % (www, zipfile))
		#     os.system('mv %s %s' % (zipfile[:-4], DATA_DIR))
		#     os.system('rm %s' % (zipfile))

def viz_pointclouds(pointcloud1, pointcloud2, association):
		pointcloud1 = pointcloud1[association,:]
		pt_indx = np.random.randint(0,pointcloud1.shape[0],1)
		diff = pointcloud1-pointcloud1[pt_indx,...]
		euclid_dist = LA.norm(diff, axis=1)
		min_val, max_val = min(euclid_dist), max(euclid_dist)
		cmap = mpl.cm.viridis
		norm = mpl.colors.Normalize(vmin=min_val, vmax=max_val)
		color_list = cmap(euclid_dist)

		# for ind, pc in enumerate(pointclouds):
		pc_o3d1 = o3d.geometry.PointCloud()
		pc_o3d1.points = o3d.utility.Vector3dVector(pointcloud1)
		pc_o3d1.colors = o3d.utility.Vector3dVector(color_list[:, :3])

		pc_o3d2 = o3d.geometry.PointCloud()
		pc_o3d2.points = o3d.utility.Vector3dVector(pointcloud2)
		pc_o3d2.colors = o3d.utility.Vector3dVector(color_list[:, :3])

		o3d.visualization.draw_geometries([pc_o3d1,pc_o3d2])

def load_data(partition, DATA_DIR, folder):
		# download(DATA_DIR)
		# BASE_DIR = os.path.dirname(os.path.abspath(__file__))
		# DATA_DIR = os.path.join(BASE_DIR, 'data')
		all_data = []
		all_label = []
		# print(DATA_DIR)
		# print(glob.glob(os.path.join(DATA_DIR, 'modelnet40_ply_hdf5_2048', 'ply_data_%s*.h5'%partition)))
		for h5_name in glob.glob(os.path.join(DATA_DIR, folder, '*%s*.h5'%partition)):
				f = h5py.File(h5_name)
				data = f['data'][:].astype('float32')
				label = f['label'][:].astype('int64')
				f.close()
				all_data.append(data)
				all_label.append(label)
		print(len(all_data))
		all_data = np.concatenate(all_data, axis=0)
		all_label = np.concatenate(all_label, axis=0)
		return all_data, all_label


def translate_pointcloud(pointcloud):
		xyz1 = np.random.uniform(low=2./3., high=3./2., size=[3])
		xyz2 = np.random.uniform(low=-0.2, high=0.2, size=[3])

		translated_pointcloud = np.add(np.multiply(pointcloud, xyz1), xyz2).astype('float32')
		return translated_pointcloud


def jitter_pointcloud(pointcloud, sigma=0.01):
		clip=2*sigma
		N, C = pointcloud.shape
		pointcloud += np.clip(sigma * np.random.randn(N, C), -1*clip, clip)
		return pointcloud


class Data3D(Dataset):
		def __init__(self, args, partition='train'):
				self.args =args
				folders = {"modelnet40": "modelnet40_hdf5_2048",
									 "modelnet10": "modelnet10_hdf5_2048",
									 "shapenetcore": "shapenetcorev2_hdf5_2048",
									 "shapenetpart": "shapenetpart_hdf5_2048",
									 "si3d": "indoor3d_sem_seg_hdf5_data",
									 "scanobjectnn": "ScanObjectNN"}
				if args.dataset=="scanobjectnn":
					print("loading scanobjectnn")
					self.data, self.label = load_ScanObjectNN(partition)
				else:
					self.data, self.label = load_data(partition, args.data_dir, folders[args.dataset])
				self.num_points = args.num_points
				self.partition = partition

		def __getitem__(self, item):
				pointcloud = self.data[item][:self.num_points]
				label = self.label[item]
				if self.partition == 'train' and self.args.w_aug:
						pointcloud = translate_pointcloud(pointcloud)
						np.random.shuffle(pointcloud)
				return pointcloud, label

		def __len__(self):
				return self.data.shape[0]

def cartesian(arrays, out=None):

		arrays = [np.asarray(x) for x in arrays]
		dtype = arrays[0].dtype

		n = np.prod([x.size for x in arrays])
		if out is None:
				out = np.zeros([n, len(arrays)], dtype=dtype)

		#m = n / arrays[0].size
		m = int(n / arrays[0].size)
		out[:,0] = np.repeat(arrays[0], m)
		if arrays[1:]:
				cartesian(arrays[1:], out=out[0:m, 1:])
				for j in range(1, arrays[0].size):
				#for j in xrange(1, arrays[0].size):
						out[j*m:(j+1)*m, 1:] = out[0:m, 1:]
		return out

def rot_mat(roll, pitch, yaw):
		# inputs: theta_rad, angle in radians
		# outputs: rotation matrix around z
		mat = euler2mat(roll, pitch, yaw)
		return np.array(mat)


def position_augmentation(pointcloud, translation, scale, angle):
		##rotation augmentation
		# angle = np.deg2rad(angle)
		rot = rot_mat(angle[0], angle[1], angle[2])
		pointcloud = np.matmul(rot,pointcloud.T).T

		##translation and scale augmentation
		translated_pointcloud = np.add(np.multiply(pointcloud, scale), translation).astype('float32')

		return translated_pointcloud

def crop_n(pc_all, pos, percent=0.1):
		dist = LA.norm(pc_all-pos,axis=1)
		top_n = np.int0((1-percent)*pc_all.shape[0])
		inds = np.argpartition(dist, -top_n)[-top_n:]
		return pc_all[inds,:]

def closest_point(nodes1, nodes2):
		nodes1 = np.asarray(nodes1)
		nodes2 = np.asarray(nodes2)
		dist_2 = np.sum((nodes1[:,np.newaxis,:] - nodes2[np.newaxis,:,:])**2, axis=-1)
		return np.argmin(dist_2, axis=0)

def load_ScanObjectNN(partition):
		BASE_DIR = 'data/ScanObjectNN'
		DATA_DIR = os.path.join(BASE_DIR, 'main_split')
		h5_name = os.path.join(DATA_DIR, f'{partition}.h5')
		f = h5py.File(h5_name)
		data = f['data'][:].astype('float32')
		label = f['label'][:].astype('int64')
		label = np.array(label)
		label = np.expand_dims(label, axis=1)
		return data, label

class Data3DUnsupervised(Dataset):
		def __init__(self, args, partition='train'):
				self.args =args
				folders = {"modelnet40": "modelnet40_hdf5_2048",
									 "modelnet10": "modelnet10_hdf5_2048",
									 "shapenetcore": "shapenetcorev2_hdf5_2048",
									 "shapenetpart": "shapenetpart_hdf5_2048",
									 "si3d": "indoor3d_sem_seg_hdf5_data"}
				self.data, self.label = load_data(partition, args.data_dir, folders[args.dataset])
				self.num_points = args.num_points
				self.partition = partition

				self.no_of_trans = self.args.no_augs
				self.no_of_angles = self.args.no_augs
				self.no_of_scalings = self.args.no_augs

				self.angle_min = -np.pi
				self.angle_max = np.pi
				self.angles = np.random.uniform(low=self.angle_min, high=self.angle_max, size=[self.no_of_angles, 3])

				self.scale_min = 2./3.
				self.scale_max = 3./2.
				self.scalings = np.random.uniform(low=self.scale_min, high=self.scale_max, size=[self.no_of_scalings, 3])

				self.trans_min = -1.
				self.trans_max = 1.
				self.translations = np.random.uniform(low=self.trans_min, high=self.trans_max, size=[self.no_of_trans,3]) # different scales
				self.combs = cartesian([np.arange(self.no_of_scalings), np.arange(self.no_of_trans), np.arange(self.no_of_angles)])
				self.device = torch.device("cuda" if args.cuda else "cpu")
				if self.args.rl_based_search:
					self.aug_mem_bank = torch.zeros((1,9)).to(self.device)
					self.dirac_states()


		def dirac_states(self):
				self.reset_augs()
				np.random.shuffle(self.scalings)
				np.random.shuffle(self.translations)
				np.random.shuffle(self.angles)

				# ind1 = self.combs[:,0]
				# ind2 = self.combs[:,1]
				# ind3 = self.combs[:,2]
				search_space = np.hstack((self.scalings, self.translations, self.angles))
				self.search_space = np.array(search_space).astype(np.float32)
				normalised_search_space = copy.deepcopy(self.search_space) ## for distance computation
				normalised_search_space[:,0:3] = (normalised_search_space[:,0:3]-self.scale_min)/(self.scale_max-self.scale_min) # scale normalized
				normalised_search_space[:,3:6] = (normalised_search_space[:,3:6]-self.trans_min)/(self.trans_max-self.trans_min) # translation normalized
				normalised_search_space[:,6:9] = (normalised_search_space[:,6:9]-self.angle_min)/(self.angle_max-self.angle_min) # translation normalized

				norm_ss = normalised_search_space
				norm_ss = torch.from_numpy(norm_ss).to(self.device)
				rl_inds, chosen_aug = dirac_sampling(self.aug_mem_bank, norm_ss)
				self.aug_mem_bank = torch.vstack((self.aug_mem_bank, chosen_aug))
				# print(rl_ind)
				# print(self.aug_mem_bank.shape)
				self.rl_scale_ind = rl_inds[0]
				self.rl_trans_ind = rl_inds[1]
				self.rl_angle_ind = rl_inds[2]

		def reset_mem(self):
					self.aug_mem_bank = torch.zeros((1,9)).to(self.device)

		def reset_augs(self):
				# print("reinitializing augmentations")
				## every epoch reinitialize the augmentations
				self.angles = np.random.uniform(low=-np.pi, high=np.pi, size=[self.no_of_angles, 3])
				self.scalings = np.random.uniform(low=2./3., high=3./2., size=[self.no_of_scalings, 3])
				self.translations = np.random.uniform(low=-1, high=1, size=[self.no_of_trans,3]) #

		def __getitem__(self, item):
				if self.args.rl_based_search and self.partition=="train":
					self.dirac_states()
					scale = self.search_space[self.rl_scale_ind,0:3]
					translation = self.search_space[self.rl_trans_ind,3:6]
					angle = self.search_space[self.rl_angle_ind,6:9]
				else:
					scale_ind = np.random.choice(self.no_of_scalings, size=1, replace=False)[0]
					scale = self.scalings[scale_ind,:]
					trans_ind = np.random.choice(self.no_of_trans, size=1, replace=False)[0]
					translation = self.translations[trans_ind,:]
					ang_ind = np.random.choice(self.no_of_angles, size=1, replace=False)[0]
					angle = self.angles[ang_ind,:]

				### before augmentations
				pointcloud1 = self.data[item][:self.num_points]
				mean = np.mean(pointcloud1, axis=0)
				pointcloud1-=mean
				if self.args.w_cropping:
					pointcloud_crop = self.data[item]
					point_ind = np.random.choice(pointcloud_crop.shape[0], size=1, replace=False)[0]
					pointcloud_crop = crop_n(pointcloud_crop, pointcloud_crop[point_ind, :], self.args.crop_ratio) # remove 10% of the pointcloud by distance.
					if pointcloud_crop.shape[0]>self.num_points:
						pointcloud_crop = pointcloud_crop[:self.num_points]
					else:
						dummy_rep = np.repeat(pointcloud_crop[[-1],:],self.num_points-pointcloud_crop.shape[0], axis=0)
						pointcloud_crop = np.vstack((pointcloud_crop, dummy_rep))
						pointcloud_crop = pointcloud_crop[:self.num_points]
					np.random.shuffle(pointcloud1)
					association = closest_point(pointcloud1.copy(), pointcloud_crop.copy()) #from pointcloud1 to pointcloud2
					pointcloud2 = position_augmentation(pointcloud_crop.copy(), translation, scale, angle) #translation, scaling and rotation augmentation

				else:
					np.random.shuffle(pointcloud1)
					# association = closest_point(pointcloud1.copy(), pointcloud1.copy()) #from pointcloud1 to pointcloud2
					association = np.arange(pointcloud1.shape[0])# pointcloud1 index to pointcloud2 index
					pointcloud2 = position_augmentation(pointcloud1.copy(), translation, scale, angle) #translation, scaling and rotation augmentation

				# viz_pointclouds(pointcloud1, pointcloud2, association)
				### after augmentations
				pointcloud1 = jitter_pointcloud(pointcloud1, sigma=self.args.jitter_var)
				pointcloud2 = jitter_pointcloud(pointcloud2, sigma=self.args.jitter_var)
				return pointcloud1, pointcloud2, association

		def __len__(self):
				# return self.combinations.shape[0]
				return self.data.shape[0]


if __name__ == '__main__':
		train = ModelNet40Unsupervised(1024)
		test = ModelNet40Unsupervised(1024, 'test')
		for data, label in train:
				print(data.shape)
				print(label.shape)
