# !/usr/bin/python

import numpy as np
import open3d as o3d
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from numpy import linalg as LA
import scipy.spatial as sp


def getYaw(kpts):
	rwb = np.mean(kpts[[1,2],:],axis=0, keepdims=True)
	fwb = np.mean(kpts[[0,3],:],axis=0, keepdims=True)

	forward_ax = (fwb-rwb)/np.linalg.norm(fwb-rwb, axis=1)
	yaw = np.arctan2(forward_ax[0,1],forward_ax[0,0])
	return yaw

def planar_plot(ax, points, colors, pt_indx, x, y):
	ax.scatter(points[:, x], points[:, y], s=5, cmap='YlGnBu', c=colors)
	ax.scatter(points[pt_indx, x], points[pt_indx, y], s=50, color='red')

def viz_feats(points, feats, substr):
	pt_indx = np.random.randint(0,points.shape[0],1)
	### euclidean distance
	diff = feats-feats[pt_indx,...]
	colors_euclid = LA.norm(diff, axis=1)

	### cosine distance
	colors_cosine = sp.distance.cdist(feats, feats[pt_indx,...], 'cosine')

	# print(points.shape, feats.shape, pt_indx, colors.shape, diff.shape)
	##normalize the data
	points=points-np.min(points,axis=0)/(np.max(points,axis=0)-np.min(points,axis=0))

	fig, axs = plt.subplots(3,2)
	fig.set_size_inches(8.0, 8.0)

	axs[0,0].set_axis_off()
	axs[1,0].set_axis_off()
	axs[2,0].set_axis_off()
	axs[0,1].set_axis_off()
	axs[1,1].set_axis_off()
	axs[2,1].set_axis_off()

	#### euclid distance
	## top view
	planar_plot(axs[0,0], points, colors_euclid, pt_indx, 0, 2)
	## side view
	planar_plot(axs[1,0], points, colors_euclid, pt_indx, 0, 1)
	## front view
	planar_plot(axs[2,0], points, colors_euclid, pt_indx, 2, 1)

	#### cosine distance
	## top view
	planar_plot(axs[0,1], points, colors_cosine, pt_indx, 0, 2)
	## side view
	planar_plot(axs[1,1], points, colors_cosine, pt_indx, 0, 1)
	## front view
	planar_plot(axs[2,1], points, colors_cosine, pt_indx, 2, 1)

	# plt.show()
	out_name = "feat_viz/"+substr+".png"
	# print("saving: "+ out_name)
	fig.savefig(out_name)
	plt.close('all')


def visualize(kpts, kptsgt, data, mean):
	kpts = kpts.cpu().detach().numpy()
	kptsgt = kptsgt.cpu().detach().numpy()
	data = data.cpu().detach().numpy()
	data = data.T
	# data = np.load(file_name).astype('float32')[:,0:3]
	# data -=mean.cpu().detach().numpy()
	# print(data.shape)
	# print(kpts.shape, data.shape)
	# kpts = np.flip(kpts, axis=0) ##reverse the order to match the order of pred kpts
	kpts = np.asarray(kpts, dtype=np.float32)[[1,3,5,7],0:3]
	kptsgt = np.asarray(kptsgt, dtype=np.float32)[[1,3,5,7],0:3]
	yaw = getYaw(kpts)
	# print("=>orientation (rad):", yaw*180/np.pi)
	## sample pointcloud from the surface of the mesh
	pcd = o3d.geometry.PointCloud()
	pcd.points = o3d.utility.Vector3dVector(data)

	## prediction kpts
	pred_kpts_pcd=[]
	for kpt in kpts:
			mesh = o3d.geometry.TriangleMesh.create_coordinate_frame()
			R = mesh.get_rotation_matrix_from_xyz((0,0,yaw))
			center = (kpt[0], kpt[1], kpt[2])
			mesh = mesh.translate(center)
			mesh = mesh.rotate(R)
			mesh = mesh.scale(0.5, center=center)
			# mesh_sphere = o3d.geometry.TriangleMesh.create_coordinate_frame(origin=kpt, size=0.8)
			pred_kpts_pcd.append(mesh)

	## gt kpts
	gt_kpts_pcd=[]
	for kpt in kptsgt:
			mesh = o3d.geometry.TriangleMesh.create_coordinate_frame()
			R = mesh.get_rotation_matrix_from_xyz((0,0,yaw))
			center = (kpt[0], kpt[1], kpt[2])
			mesh = mesh.translate(center)
			mesh = mesh.rotate(R)
			mesh = mesh.scale(0.5, center=center)
			# mesh_sphere = o3d.geometry.TriangleMesh.create_coordinate_frame(origin=kpt, size=0.8)
			gt_kpts_pcd.append(mesh)

	o3d.visualization.draw_geometries([pcd]+pred_kpts_pcd+gt_kpts_pcd)
