#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Fine tuning with supervised approach to see if there is any improvement, compared to supervised only approach
"""

from __future__ import print_function
import os
import argparse
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from torch.optim.lr_scheduler import CosineAnnealingLR
import numpy as np
from torch.utils.data import DataLoader
import sklearn.metrics as metrics
from tqdm import tqdm
from config.config import *
import sklearn.metrics as metrics
from lib.visualize import viz_feats
from lib.model import PointNet, DGCNN, DGCNN_partseg
from lib.util import cal_loss, IOStream, cosine_distance, RMSELoss
from lib.dataloaders import Data3D, Data3DUnsupervised
from sklearn.svm import SVC
import pickle
import joblib
# load args
args = parse_args()

device = torch.device("cuda" if args.cuda else "cpu")

### setup dataloader
if args.unsupervised:
		dataset = Data3DUnsupervised
else:
		dataset = Data3D

### setup model
if args.model == 'pointnet':
		model = PointNet(args).to(device)
elif args.model == 'dgcnn':
		if args.task =="classification":
			model = DGCNN(args).to(device)
		if args.task =='segmentation':
			model = DGCNN_partseg(args).to(device)
else:
		raise Exception("Not implemented")

# if args.rl_based_search:
# 	# rl_model = ValueNet(args).to(device)
# 	rmse_loss = RMSELoss()
# else:
rl_model = None

def _init_(args):
		# copying models and files to corresponding checkpoints
		if not os.path.exists('checkpoints'):
				os.makedirs('checkpoints')
		if not os.path.exists('checkpoints/'+args.exp_name):
				os.makedirs('checkpoints/'+args.exp_name)
		if not os.path.exists('checkpoints/'+args.exp_name+'/'+'models'):
				os.makedirs('checkpoints/'+args.exp_name+'/'+'models')
		if not os.path.exists(args.output_folder):
				os.makedirs(args.output_folder)
		# os.system('cp main.py checkpoints'+'/'+args.exp_name+'/'+'main.py.backup')
		# os.system('cp model.py checkpoints' + '/' + args.exp_name + '/' + 'model.py.backup')
		# os.system('cp util.py checkpoints' + '/' + args.exp_name + '/' + 'util.py.backup')
		# os.system('cp dataloaders.py checkpoints' + '/' + args.exp_name + '/' + 'data.py.backup')
def train_val_gen(loader, model, epoch=0, opt=None, phase="test", viz_feats=False):
		# train-val method for supervised approach
		train_loss = 0.0
		count = 0.0
		model.eval()
		all_feats = []
		all_labels = []
		# load supervised data loader
		for indx, (data, label) in enumerate(tqdm(loader)):
				labels = list(map(lambda x: x[0],label.numpy().tolist()))
				data = data.to(device)
				data = data.permute(0, 2, 1)
				batch_size = data.size()[0]
				opt.zero_grad()
				feats = model(data)[2]
				feats = feats.detach().cpu().numpy()
				label = label.detach().cpu().numpy()
				all_feats.append(feats)
				all_labels+=labels
				# print(feats.shape, label.shape)
		all_feats = np.vstack(all_feats)
		all_labels= np.array(all_labels)
		return all_feats, all_labels

def test_svm(X_test, y_test, filename):
		# load the model from disk
		svm = joblib.load(filename.replace("model.t7", "model_svm.sav"))
		return svm.score(X_test, y_test)

def train_svm(X_train, y_train, X_test, y_test):
		svm = SVC(C = 0.1, kernel ='linear')
		svm.fit(X_train, y_train)
		return svm.score(X_test, y_test), svm

def train(args, io):
		# train method abstraction for both supervised and unsupervised

		global dataset, model

		all_params = []
		args.w_aug = False
		train_dataset = dataset(partition='train', args=args)
		train_loader = DataLoader(train_dataset, num_workers=0,
															batch_size=args.batch_size, shuffle=True, drop_last=True)
		val_dataset = dataset(partition='test', args=args)
		val_loader = DataLoader(val_dataset, num_workers=0,
															batch_size=args.batch_size, shuffle=False, drop_last=True)

		print(str(model))

		model = nn.DataParallel(model)
		model.load_state_dict(torch.load(args.model_path))
		all_params+=model.parameters()

		print("Let's use", torch.cuda.device_count(), "GPUs!")

		if args.use_sgd:
				print("Use SGD")
				opt = optim.SGD(all_params, lr=args.lr*100, momentum=args.momentum, weight_decay=1e-4)
		else:
				print("Use Adam")
				opt = optim.Adam(all_params, lr=args.lr, weight_decay=1e-4)

		# according to dgcnn model for sgd
		scheduler = CosineAnnealingLR(opt, args.epochs, eta_min=args.lr)

		best_val_loss = float('inf')

		## generate features and labels for svm
		epoch=0
		train_feats, train_labels = train_val_gen(train_loader, model, epoch, opt, phase="train", viz_feats=args.viz_feats)
		print(train_feats.shape, train_labels.shape)
		val_feats, val_labels = train_val_gen(val_loader, model, epoch, opt, phase="val", viz_feats=args.viz_feats)
		# print(val_feats.shape, val_labels.shape)
		# with open('data/svm_data.pkl','wb') as f:
		# 	data = {"train_feats":train_feats, "train_labels":train_labels, "val_feats":val_feats, "val_labels":val_labels}
		# 	pickle.dump(data, f)

		# with open('data/svm_data.pkl','rb') as f:
		# 	data = pickle.load(f)
		# 	train_feats = data["train_feats"]
		# 	train_labels = data["train_labels"]
		# 	val_feats = data["val_feats"]
		# 	val_labels = data["val_labels"]

		score, _ = train_svm(train_feats, train_labels, val_feats, val_labels)
		print("score:", score)
		score = test_svm(val_feats, val_labels, args.model_path)
		print("score:", score)

if __name__ == "__main__":

		_init_(args)

		io = IOStream('checkpoints/' + args.exp_name + '/run.log')
		io.cprint(str(args))

		# with open(args.output_folder+'/'+args.exp_name+'_train.txt','w+') as f:
		# 	f.write("epoch, learning rate, loss, overall accuracy, mean class accuracy\n")

		# with open(args.output_folder+'/'+args.exp_name+'_test.txt','w+') as f:
		# 	f.write("epoch, learning rate, loss, overall accuracy, mean class accuracy\n")

		# with open(args.output_folder+'/'+args.exp_name+'_val.txt','w+') as f:
		# 	f.write("epoch, learning rate, loss, overall accuracy, mean class accuracy\n")

		args.cuda = args.cuda and torch.cuda.is_available()
		torch.manual_seed(args.seed)
		if args.cuda:
				io.cprint(
						'Using GPU : ' + str(torch.cuda.current_device()) + ' from ' + str(torch.cuda.device_count()) + ' devices')
				torch.cuda.manual_seed(args.seed)
		else:
				io.cprint('Using CPU')

		if not args.eval:
				train(args, io)
		else:
				test(args, io)
